import copy
import re
import threading
import timeit

import calibrationBase
import gpu_utils
import hashToSchema
import numpy as np
import pycuda.compiler
import pycuda.driver
import pycuda.gpuarray
import pycuda.tools
import utils
from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    INPUT_CHANNEL,
    INT32_ELEMENT,
    KARABO_CLASSINFO,
    NDARRAY_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    SLOT_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    VECTOR_STRING_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    ChannelMetaData,
    Epochstamp,
    Hash,
    MetricPrefix,
    Schema,
    Timestamp,
    Trainstamp,
    Unit,
)
from karabo.common.states import State

from .cuda_pipeline import PyCudaPipeline


@KARABO_CLASSINFO("DsscCorrection", "0.0.0")
class DsscCorrection(calibrationBase.CalibrationReceiverBaseDevice):
    _dict_cache_slots = {
        "applyCorrection",
        "doAnything",
        "dataFormat.memoryCells",
        "dataFormat.memoryCellsCorrection",
        "dataFormat.pixelsX",
        "dataFormat.pixelsY",
        "preview.enable",
        "preview.pulse",
        "preview.trainIdModulo",
        "processingStateTimeout",
        "performance.rateUpdateOnEachInput",
        "state",
    }

    @staticmethod
    def expectedParameters(expected):
        DsscCorrection.addConstant(
            "Offset", "Dark", expected, optional=True, mandatoryForIteration=True
        )

        SLOT_ELEMENT(expected).key("askConnectedReadersToSendMySources").displayedName(
            "Request sources from connected RunToPipe"
        ).description(
            "Only relevant for development environment. When running without a "
            "CAL_MANAGER, we need to tell RunToPipe instances which sources "
            "to send us."
        ).commit()

        BOOL_ELEMENT(expected).key("doAnything").displayedName(
            "Enable input processing"
        ).description(
            "Toggle handling of input (at all). If False, the input handler "
            "of this device will be skipped. Useful to decrease logspam if "
            "device is misconfigured."
        ).assignmentOptional().defaultValue(
            True
        ).reconfigurable().commit()

        BOOL_ELEMENT(expected).key("applyCorrection").displayedName(
            "Enable correction(s)"
        ).description(
            "Toggle whether or not correction(s) are applied to image data. "
            "If false, this device still reshapes data to output shape, "
            "applies the pulse filter, and casts to output dtype. Useful for "
            "inspecting the raw data in the same format as corrected data."
        ).assignmentOptional().defaultValue(
            True
        ).reconfigurable().commit()

        INPUT_CHANNEL(expected).key("dataInput").commit()
        # note: output schema not set, will be updated to match data later
        OUTPUT_CHANNEL(expected).key("dataOutput").commit()

        VECTOR_STRING_ELEMENT(expected).key("fastSources").displayedName(
            "Fast data sources"
        ).description(
            "Sources to fast data as provided in channel metadata. "
            "Provide in the form source@path.in.hash to identify both source "
            "and path in the source data hash.\n"
            "Currently ignores the path.in.hash part (for hardcoded image.data)"
        ).assignmentMandatory().commit()

        STRING_ELEMENT(expected).key("pulseFilter").displayedName(
            "Pulse filter"
        ).description(
            "Filter pulses: will be evaluated as array of indices to keep from data. "
            "Can be anything which can be turned into numpy uint16 array. "
            "Numpy is available as np. "
            "Take care not to include duplicates. "
            "If empty, will not filter at all."
        ).assignmentOptional().defaultValue(
            ""
        ).reconfigurable().commit()

        NODE_ELEMENT(expected).key("dataFormat").displayedName(
            "Data format (in/out)"
        ).commit()
        STRING_ELEMENT(expected).key("dataFormat.inputImageDtype").displayedName(
            "Input image data dtype"
        ).description("The (numpy) dtype to expect for incoming image data.").options(
            "uint16,float32"
        ).assignmentOptional().defaultValue(
            "uint16"
        ).commit()
        STRING_ELEMENT(expected).key("dataFormat.outputImageDtype").displayedName(
            "Output image data dtype"
        ).description(
            "The (numpy) dtype to use for outgoing image data. "
            "Input is cast to float32, corrections are applied, and only then "
            "will the result be cast back to outputImageDtype (all on GPU)."
        ).options(
            "float16,float32,uint16"
        ).assignmentOptional().defaultValue(
            "float32"
        ).commit()
        # important: shape of data as going into correction
        UINT32_ELEMENT(expected).key("dataFormat.pixelsX").displayedName(
            "Pixels x"
        ).description(
            "Number of pixels of image data along X axis"
        ).assignmentMandatory().commit()
        UINT32_ELEMENT(expected).key("dataFormat.pixelsY").displayedName(
            "Pixels y"
        ).description(
            "Number of pixels of image data along Y axis"
        ).assignmentMandatory().commit()
        UINT32_ELEMENT(expected).key("dataFormat.memoryCells").displayedName(
            "Memory cells"
        ).description(
            "Full number of memory cells in incoming data"
        ).assignmentMandatory().commit()
        UINT32_ELEMENT(expected).key("dataFormat.memoryCellsCorrection").displayedName(
            "(Debug) Memory cells in correction map"
        ).description(
            "Full number of memory cells in currently loaded correction map. "
            "May exceed memory cell number in input if veto is on. "
            "This value just displayed for debugging."
        ).readOnly().initialValue(
            0
        ).commit()
        VECTOR_UINT32_ELEMENT(expected).key("dataFormat.inputDataShape").displayedName(
            "Input data shape"
        ).description(
            "Image data shape in incoming data (from reader / DAQ). "
            "Value computed from pixelsX, pixelsY, and memoryCells - "
            "this slot is just showing you what is currently expected."
        ).readOnly().initialValue(
            []
        ).commit()
        VECTOR_UINT32_ELEMENT(expected).key("dataFormat.outputDataShape").displayedName(
            "Output data shape"
        ).description(
            "Image data shape for data output from this device. "
            "Value computed from pixelsX, pixelsY, and the size of the pulse filter - "
            "this slot is just showing what is currently expected."
        ).readOnly().initialValue(
            []
        ).commit()

        UINT32_ELEMENT(expected).key("outputShmemBufferLength").displayedName(
            "Output buffer length"
        ).description(
            "Corrected trains are written to shared memory locations. These are "
            "pre-allocated and re-used. This parameter determines how big "
            "(number of trains) the circular buffer will be."
        ).assignmentOptional().defaultValue(
            50
        ).commit()

        # preview schema (WIP)
        NODE_ELEMENT(expected).key("preview").displayedName("Preview").commit()
        preview_schema = Schema()
        NODE_ELEMENT(preview_schema).key("data").commit()
        NDARRAY_ELEMENT(preview_schema).key("data.adc").dtype("FLOAT").commit()
        OUTPUT_CHANNEL(expected).key("preview.outputRaw").dataSchema(
            preview_schema
        ).commit()
        OUTPUT_CHANNEL(expected).key("preview.outputCorrected").dataSchema(
            preview_schema
        ).commit()
        BOOL_ELEMENT(expected).key("preview.enable").displayedName(
            "Enable preview data generation"
        ).assignmentOptional().defaultValue(True).reconfigurable().commit()
        INT32_ELEMENT(expected).key("preview.pulse").displayedName(
            "Pulse (or stat) for preview"
        ).description(
            "If this value is ≥ 0, the corresponding index from data will be "
            "sliced for the preview. If this value is ≤ 0, preview will be one "
            "of the following stats:\n"
            "-1: max\n"
            "-2: mean\n"
            "-3: sum\n"
            "-4: stdev\n"
            "Max means selecting the pulse with the maximum integrated value. "
            "The others are computed across all filtered pulses in the train."
        ).assignmentOptional().defaultValue(
            0
        ).reconfigurable().commit()
        UINT32_ELEMENT(expected).key("preview.trainIdModulo").displayedName(
            "Train modulo for throttling"
        ).description(
            "Preview will only be sent for trains whose ID modulo this number "
            "is zero. Higher values means fewer preview updates. Should be "
            "adjusted based on input rate. Keep in mind that the GUI has "
            "limited refresh rate anyway and that network is precious."
        ).assignmentOptional().defaultValue(
            6
        ).reconfigurable().commit()

        # timer-related settings
        NODE_ELEMENT(expected).key("performance").displayedName(
            "Performance measures"
        ).commit()
        FLOAT_ELEMENT(expected).key("performance.rateUpdateInterval").displayedName(
            "Rate update interval"
        ).description(
            "Maximum interval (seconds) between updates of the rate. "
            "Mostly relevant if not rateUpdateOnEachInput or if input is slow."
        ).assignmentOptional().defaultValue(
            1
        ).reconfigurable().commit()
        FLOAT_ELEMENT(expected).key("performance.rateBufferSpan").displayedName(
            "Rate measurement buffer span"
        ).description(
            "Event buffer timespan (in seconds) for measuring rate"
        ).assignmentOptional().defaultValue(
            20
        ).reconfigurable().commit()
        BOOL_ELEMENT(expected).key("performance.rateUpdateOnEachInput").displayedName(
            "Update rate on each input"
        ).description(
            "Whether or not to update the device rate for each input "
            "(otherwise only based on rateUpdateInterval). "
            "Note that processed trains are always registered - this just "
            "impacts when the rate is computed based on this."
        ).assignmentOptional().defaultValue(
            False
        ).reconfigurable().commit()

        FLOAT_ELEMENT(expected).key("processingStateTimeout").description(
            "Timeout after which the device goes from PROCESSING back to ACTIVE "
            "if no new input is processed"
        ).assignmentOptional().defaultValue(10).reconfigurable().commit()

        # just measurements and counters to display
        UINT64_ELEMENT(expected).key("trainId").displayedName("Train ID").description(
            "ID of latest train processed by this device."
        ).readOnly().initialValue(0).commit()
        FLOAT_ELEMENT(expected).key("performance.lastProcessingDuration").displayedName(
            "Processing time"
        ).description(
            "Amount of time spent in processing latest train. "
            "Time includes generating preview and sending data."
        ).unit(
            Unit.SECOND
        ).metricPrefix(
            MetricPrefix.MILLI
        ).readOnly().initialValue(
            0
        ).commit()
        FLOAT_ELEMENT(expected).key("performance.rate").displayedName(
            "Rate"
        ).description(
            "Actual rate with which this device gets / processes / sends trains"
        ).unit(
            Unit.HERTZ
        ).readOnly().initialValue(
            0
        ).commit()
        FLOAT_ELEMENT(expected).key("performance.theoreticalRate").displayedName(
            "Processing rate (hypothetical)"
        ).description(
            "Rate with which this device could hypothetically process trains. "
            "Based on lastProcessingDuration."
        ).unit(
            Unit.HERTZ
        ).readOnly().initialValue(
            float("NaN")
        ).warnLow(
            10
        ).info(
            "Processing not fast enough for full speed"
        ).needsAcknowledging(
            False
        ).commit()

        # stuff from typical calPy that we don't use right now
        # Included to avoid errors due to unexpected configuration from init device
        STRING_ELEMENT(expected).key("sourceInfix").displayedName(
            "[Disabled]"
        ).assignmentOptional().defaultValue("").commit()

        UINT32_ELEMENT(expected).key("maxGPUHandlesInFlight").displayedName(
            "[Disabled]"
        ).assignmentOptional().defaultValue(35).commit()

        VECTOR_UINT32_ELEMENT(expected).key("gainMapping").displayedName(
            "[Disabled]"
        ).assignmentOptional().defaultValue([]).commit()

        BOOL_ELEMENT(expected).key("dontProcess").displayedName(
            "[Disabled]"
        ).assignmentOptional().defaultValue(False).reconfigurable().commit()

    def __init__(self, config):
        self._dict_cache = {k: config.get(k) for k in self._dict_cache_slots}
        super().__init__(config)
        pycuda.driver.init()
        self.gpu_context = pycuda.tools.make_default_context()

        # very sneaky debugging
        self.KARABO_SLOT(self.askConnectedReadersToSendMySources)

        self.KARABO_ON_DATA("dataInput", self.process_input)
        self.KARABO_ON_EOS("dataInput", self.handle_eos)

        self.sources = set(config.get("fastSources"))

        self.input_data_dtype = getattr(np, config.get("dataFormat.inputImageDtype"))
        self.output_data_dtype = getattr(np, config.get("dataFormat.outputImageDtype"))
        self._offset_map = None
        self._update_pulse_filter(config.get("pulseFilter"))
        self._update_shapes(
            config.get("dataFormat.pixelsX"),
            config.get("dataFormat.pixelsY"),
            config.get("dataFormat.memoryCells"),
            self.pulse_filter,
        )
        self._has_set_output_schema = False
        self._has_set_preview_output_schema = False
        self._rate_tracker = calibrationBase.utils.UpdateRate(
            interval=config.get("performance.rateBufferSpan")
        )
        self.managerInstance = None
        self.KARABO_SLOT(self.registerManager)
        self._reset_timer = None
        self.updateState(State.ON)

        self._buffered_status_update = Hash(
            "trainId",
            0,
            "performance.rate",
            0,
            "performance.theoreticalRate",
            float("NaN"),
            "performance.lastProcessingDuration",
            0,
        )
        self._rate_update_timer = utils.RepeatingTimer(
            interval=config.get("performance.rateUpdateInterval"),
            callback=self._update_actual_rate,
        )
        self._buffer_lock = threading.Lock()

    def get(self, key):
        if key in self._dict_cache_slots:
            return self._dict_cache.get(key)
        else:
            return super().get(key)

    def set(self, *args):
        if len(args) == 2:
            key, value = args
            if key in self._dict_cache_slots:
                self._dict_cache[key] = value
        super().set(*args)

    def __del__(self):
        self.gpu_context.detach()

    def preReconfigure(self, config):
        if config.has("pulseFilter"):
            with self._buffer_lock:
                # apply new pulse filter
                self._update_pulse_filter(config.get("pulseFilter"))
                # but existing shapes (not reconfigurable)
                # TODO: avoid double compilation here if constants are loaded
                self._update_shapes(
                    self.get("dataFormat.pixelsX"),
                    self.get("dataFormat.pixelsY"),
                    self.get("dataFormat.memoryCells"),
                    self.pulse_filter,
                )

        if config.has("performance.rateUpdateInterval"):
            self._rate_update_timer.stop()
            self._rate_update_timer = utils.RepeatingTimer(
                interval=config.get("performance.rateUpdateInterval"),
                callback=self._update_actual_rate,
            )

        if config.has("performance.rateBufferSpan"):
            self._rate_tracker = calibrationBase.utils.UpdateRate(
                interval=config.get("performance.rateBufferSpan")
            )

        for path in config.getPaths():
            if path in self._dict_cache_slots:
                self._dict_cache[path] = config.get(path)

    def process_input(self, data, metadata):
        """Registered for dataInput, handles all processing and sending

        Comparable to StreamBase.onInput but hopefully faster

        """

        if not self.get("doAnything"):
            if self.get("state") is State.PROCESSING:
                self.updateState(State.ACTIVE)
            return

        # TODO: compare KARABO_ON_INPUT (old) against KARABO_ON_DATA (current)
        source = metadata.get("source")

        if source not in self.sources:
            self.log.INFO(f"Ignoring unknown source {source}")
            return

        # TODO: what are these empty things for?
        if not data.has("image"):
            self.log.INFO("Ignoring hash without image node")
            return

        time_start = timeit.default_timer()

        train_id = metadata.getAttribute("timestamp", "tid")
        cell_table = np.squeeze(data.get("image.cellId"))
        assert isinstance(cell_table, np.ndarray), "image.cellId should be ndarray"
        if len(cell_table.shape) == 0:
            msg = "cellId had 0 dimensions. DAQ may not be sending data."
            self.set("status", msg)
            self.log.WARN(msg)
            return
        # original shape: 400, 1, 128, 512 (memory cells, something, y, x)
        # TODO: consider making paths configurable
        image_data = data.get("image.data")
        if image_data.shape[0] != self.get("dataFormat.memoryCells"):
            self.set(
                "status", f"Updating input shapes based on received {image_data.shape}"
            )
            # TODO: truncate if > 800
            self.set("dataFormat.memoryCells", image_data.shape[0])
            with self._buffer_lock:
                self._update_pulse_filter(self.get("pulseFilter"))
                self._update_shapes(
                    self.get("dataFormat.pixelsX"),
                    self.get("dataFormat.pixelsY"),
                    self.get("dataFormat.memoryCells"),
                    self.pulse_filter,
                )
        # TODO: check shape (DAQ fake data and RunToPipe don't agree)
        # TODO: consider just updating shapes based on whatever comes in

        do_generate_preview = train_id % self.get(
            "preview.trainIdModulo"
        ) == 0 and self.get("preview.enable")
        do_apply_correction = self.get("applyCorrection")

        if not self.get("state") is State.PROCESSING:
            self.updateState(State.PROCESSING)
            self.set("status", "Processing data")
        if self._reset_timer is None:
            self._reset_timer = utils.DelayableTimer(
                timeout=self.get("processingStateTimeout"),
                callback=self._reset_state_from_processing,
            )
        else:
            self._reset_timer.set_timeout(self.get("processingStateTimeout"))

        with self._buffer_lock:
            cell_table = cell_table[self.pulse_filter]
            pulse_table = np.squeeze(data.get("image.pulseId"))[self.pulse_filter]

            cell_table_max = np.max(cell_table)
            correction_cell_num = self.get("dataFormat.memoryCellsCorrection")
            if do_apply_correction:
                if correction_cell_num == 0:
                    msg = "No constant loaded, correction will not be applied."
                    self.log.WARN(msg)
                    self.set("status", msg)
                    do_apply_correction = False
                elif cell_table_max >= correction_cell_num:
                    msg = (
                        f"Max cell ID ({cell_table_max}) exceeds range for loaded "
                        f"constant (has {correction_cell_num} cells). Some frames "
                        "will not be corrected."
                    )
                    self.log.WARN(msg)
                    self.set("status", msg)

            with gpu_utils.GPUContextContext(self.gpu_context):
                self.gpu_buffer_cell_table.set(cell_table)
                self.gpu_buffer_input_image_data.set(image_data)
                self.pipeline.reshape(
                    self.gpu_buffer_input_image_data,
                    self.gpu_buffer_reshaped_image_data,
                )
                if do_apply_correction:
                    buffer_handle, result = self.pipeline.correct(
                        self.gpu_buffer_reshaped_image_data,
                        self.gpu_buffer_cell_table,
                    )
                else:
                    buffer_handle, result = self.pipeline.only_cast(
                        self.gpu_buffer_reshaped_image_data
                    )
                if do_generate_preview:
                    preview_slice_index = self.get("preview.pulse")
                    if preview_slice_index >= 0:
                        # look at pulse_table to find which index this pulse ID is in
                        pulse_id_found = np.where(pulse_table == preview_slice_index)[0]
                        if len(pulse_id_found) == 0:
                            pulse_found_instead = pulse_table[0]
                            msg = (
                                f"Pulse {preview_slice_index} not found in "
                                f"image.pulseId, arbitrary pulse "
                                f"{pulse_found_instead} will be shown."
                            )
                            preview_slice_index = 0
                            self.log.WARN(msg)
                            self.set("status", msg)
                        else:
                            preview_slice_index = pulse_id_found[0]
                    preview_raw, preview_corrected = self.pipeline.compute_preview(
                        self.gpu_buffer_reshaped_image_data,
                        preview_slice_index,
                        reuse_corrected=do_apply_correction,
                        cell_table=self.gpu_buffer_cell_table,
                    )

        data.set("image.data", buffer_handle)
        data.set("image.cellId", cell_table[:, np.newaxis])
        data.set("image.pulseId", pulse_table[:, np.newaxis])
        self.write_output(data, metadata)
        if do_generate_preview:
            self.write_combiner_preview(
                preview_raw, preview_corrected, train_id, source
            )

        # update rate etc.
        self._buffered_status_update.set("trainId", train_id)
        self._rate_tracker.update()
        time_spent = timeit.default_timer() - time_start
        self._buffered_status_update.set(
            "performance.lastProcessingDuration", time_spent * 1000
        )
        if self.get("performance.rateUpdateOnEachInput"):
            self._update_actual_rate()

    def handle_eos(self, channel):
        self._has_set_output_schema = False
        self.updateState(State.ON)
        self.signalEndOfStream("dataOutput")

    def write_output(self, data, old_metadata):
        metadata = ChannelMetaData(
            old_metadata.get("source"),
            Timestamp.fromHashAttributes(old_metadata.getAttributes("timestamp")),
        )

        if "image.passport" not in data:
            data["image.passport"] = []
        data["image.passport"].append(self.getInstanceId())

        if not self._has_set_output_schema:
            self.updateState(State.CHANGING)
            self._update_output_schema(data)
            self.updateState(State.PROCESSING)

        channel = self.signalSlotable.getOutputChannel("dataOutput")
        channel.write(data, metadata, False)
        channel.update()

    def write_combiner_preview(self, data_raw, data_corrected, train_id, source):
        # TODO: take into account updated pulse table after pulse filter
        preview_hash = Hash()
        preview_hash.set("image.passport", [self.getInstanceId()])
        preview_hash.set("image.trainId", train_id)
        preview_hash.set("image.pulseId", self.get("preview.pulse"))

        # note: have to construct because setting .tid after init is broken
        timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
        metadata = ChannelMetaData(source, timestamp)
        for channel_name, data in (
            ("preview.outputRaw", data_raw),
            ("preview.outputCorrected", data_corrected),
        ):
            preview_hash.set("data.adc", data[..., np.newaxis])
            channel = self.signalSlotable.getOutputChannel(channel_name)
            channel.write(preview_hash, metadata, False)
            channel.update()

    def getConstant(self, name):
        """Hacky override of getConstant to actually return None on failure

        Full function is from CalibrationReceiverBaseDevice

        """

        const = super().getConstant(name)
        if const is not None and len(const.shape) == 1:
            self.log.WARN(
                f"Constant {name} should probably be None, but is array"
                f" of size {const.size}, shape {const.shape}"
            )
            const = None
        return const

    def constantLoaded(self):
        """Hook from CalibrationReceiverBaseDevice called after each getConstant

        Here, used to load the received constants (or correction maps derived
        fromt them) onto GPU.

        TODO: call after receiving *all* constants instead of calling once per
        new constant (will cause some overhead for bigger devices)

        """

        offset_map = self.getConstant("Offset")
        input_memory_cells = self.get("dataFormat.memoryCells")
        if offset_map is None:
            msg = (
                "Warning: Did not find offset constant, offset correction "
                "will not be applied"
            )
            self.set("status", msg)
            self.log.WARN(msg)
            self._offset_map = None
        elif len(offset_map.shape) not in (3, 4):
            msg = (
                f"Offset map had unexpected shape {offset_map.shape}, "
                "offset correction will not be applied"
            )
            self.set("status", msg)
            self.log.WARN(msg)
        else:
            self.log.INFO(f"Offset map loaded has shape {offset_map.shape}")
            if len(offset_map.shape) == 4:  # old format (see offsetcorrection_dssc.py)?
                offset_map = offset_map[..., 0]
            constant_memory_cells = offset_map.shape[-1]
            if input_memory_cells > constant_memory_cells:
                msg = (
                    f"Warning: Memory cells in input {input_memory_cells} > "
                    f"memory cells in constant {constant_memory_cells}, some "
                    "frames may not get correction applied."
                )
                self.set("status", msg)
                self.log.WARN(msg)
            self._offset_map = offset_map.astype(np.float32)
            msg = f"Offset map with shape {self._offset_map.shape} ready to load to GPU"
            self.set("status", msg)
            self.log.INFO(msg)
            if constant_memory_cells != self.get("dataFormat.memoryCellsCorrection"):
                self.log.INFO("Will first have to update buffers on GPU")
                self.set("dataFormat.memoryCellsCorrection", constant_memory_cells)

        self._update_maps_on_gpu()

    def registerManager(self, instance_id):
        """A hook from stream.py for Manager devices to register themselves

        instance_id should be the instance id of the manager device.  The
        registration is currently not really used I think.

        """

        self.managerInstance = instance_id
        self.log.INFO(f"Registered calibration manager {instance_id}")

    def askConnectedReadersToSendMySources(self):
        """For all connected outputs, set my sources to True

        This is really something the cal_manager should handle. Will be removed
        after prototyping phase.

        """

        connected_friends = [
            # TODO: full set of possible device / channel separators
            re.split(r"[@:]", channel)[0]
            for channel in self.get("dataInput.connectedOutputChannels")
        ]

        send_me_the_thing_hash = Hash()
        for source in self.get("fastSources"):
            send_me_the_thing_hash[f"sources.{source}"] = True

        # ask them nicely to send it
        for friend in connected_friends:
            # TODO: check that device is a reader
            self.signalSlotable.call(friend, "slotReconfigure", send_me_the_thing_hash)
        # who needs middlelayer? we can roll out own setNowait 😎

    def _update_output_schema(self, data):
        """Updates the schema of dataOutput based on parameter data (a Hash)

        This should only be called once: when handling output for the first
        time, we update the schema to match the modified data we'd send.

        """

        self.log.INFO("Updating the output schema based on actual outgoing data")
        my_schema = self.getFullSchema()
        data_schema = hashToSchema.HashToSchema(data).schema
        OUTPUT_CHANNEL(my_schema).key("dataOutput").dataSchema(data_schema).commit()
        my_config = copy.copy(self.getCurrentConfiguration())
        self.updateSchema(my_schema)
        self.log.INFO("Re-applying backed up config")
        self.set(my_config)
        self._has_set_output_schema = True
        self.log.INFO("Ready to continue")

    def _update_pulse_filter(self, filter_string):
        """Called whenever the pulse filter changes - typically followed by _update_shapes"""

        if filter_string.strip() == "":
            new_filter = np.arange(self.get("dataFormat.memoryCells"), dtype=np.uint16)
        else:
            new_filter = np.array(eval(filter_string), dtype=np.uint16)
        assert np.max(new_filter) < self.get("dataFormat.memoryCells")
        self.pulse_filter = new_filter

    def _update_shapes(self, pixels_x, pixels_y, memory_cells, pulse_filter):
        """(Re)initialize (GPU) buffers according to expected data shapes"""

        input_data_shape = (memory_cells, 1, pixels_y, pixels_x)
        output_data_shape = (pixels_x, pixels_y, pulse_filter.size)
        num_buffered_trains = self.get("outputShmemBufferLength")
        self.set("dataFormat.inputDataShape", list(input_data_shape))
        self.set("dataFormat.outputDataShape", list(output_data_shape))

        shmem_buffer_name = self.getInstanceId().replace("/", "_") + ":dataOutput"

        with gpu_utils.GPUContextContext(self.gpu_context):
            self.pipeline = PyCudaPipeline(
                pixels_x,
                pixels_y,
                memory_cells,
                pulse_filter,
                output_buffer_name=shmem_buffer_name,
                output_buffer_size=num_buffered_trains,
                input_data_dtype=self.input_data_dtype,
                output_data_dtype=self.output_data_dtype,
            )
            self.gpu_buffer_cell_table = pycuda.gpuarray.empty(
                pulse_filter.size, dtype=np.uint16
            )
            self.gpu_buffer_input_image_data = pycuda.gpuarray.empty(
                input_data_shape, dtype=self.input_data_dtype
            )
            self.gpu_buffer_reshaped_image_data = pycuda.gpuarray.empty(
                output_data_shape, dtype=self.input_data_dtype
            )

        self._update_maps_on_gpu()

    def _update_maps_on_gpu(self):
        """Updates the correction maps stored on GPU based on constants known

        This only does something useful if constants have been retrieved from
        CalCat.  Should be called automatically upon retrieval and after
        changing the data shape.

        """

        self.set("status", "Updating constants on GPU using known constants")
        self.updateState(State.CHANGING)
        if self._offset_map is not None:
            with gpu_utils.GPUContextContext(self.gpu_context):
                self.pipeline.load_constants(self._offset_map)
            msg = "Done transferring known constant(s) to GPU"
            self.log.INFO(msg)
            self.set("status", msg)

        self.updateState(State.ON)

    def _reset_state_from_processing(self):
        if self.get("state") is State.PROCESSING:
            self.updateState(State.ON)
            self._reset_timer = None

    def _update_actual_rate(self):
        if not self.get("state") is State.PROCESSING:
            self._rate_update_timer.delay()
            return
        self._buffered_status_update.set("performance.rate", self._rate_tracker.rate())
        last_processing = self._buffered_status_update.get(
            "performance.lastProcessingDuration"
        )
        if last_processing > 0:
            theoretical_rate = 1000 / last_processing
            self._buffered_status_update.set(
                "performance.theoreticalRate", theoretical_rate
            )
        self.set(self._buffered_status_update)
        self._rate_update_timer.delay()
